import { chain, delay, action, tween, spring, keyframes, inertia, physics, easing } from 'popmotion';
import { number, degrees, percent, px, vw, vh, complex } from 'style-value-types';
import styler, { createStylerFactory, buildStyleProperty, isTransformProp } from 'stylefire';
import sync, { getFrameData, cancelSync } from 'framesync';
import isPropValid from '@emotion/is-prop-valid';
import { useEffect, useRef, forwardRef, useMemo, memo, createContext, useContext, useState, createElement, useCallback } from 'react';
import { velocityPerSecond, mix, interpolate, wrap } from '@popmotion/popcorn';
import { invariant, warning } from 'hey-listen';

// These are identity functions that help TypeScript understand the type of the provided array
var htmlTuple = function (args) { return args; };
var svgTuple = function (args) { return args; };
/**
 * @internal
 */
var htmlElements = htmlTuple([
    "a",
    "abbr",
    "address",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "base",
    "bdi",
    "bdo",
    "big",
    "blockquote",
    "body",
    "br",
    "button",
    "canvas",
    "caption",
    "cite",
    "code",
    "col",
    "colgroup",
    "data",
    "datalist",
    "dd",
    "del",
    "details",
    "dfn",
    "dialog",
    "div",
    "dl",
    "dt",
    "em",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "head",
    "header",
    "hgroup",
    "hr",
    "html",
    "i",
    "iframe",
    "img",
    "input",
    "ins",
    "kbd",
    "keygen",
    "label",
    "legend",
    "li",
    "link",
    "main",
    "map",
    "mark",
    "menu",
    "menuitem",
    "meta",
    "meter",
    "nav",
    "noscript",
    "object",
    "ol",
    "optgroup",
    "option",
    "output",
    "p",
    "param",
    "picture",
    "pre",
    "progress",
    "q",
    "rp",
    "rt",
    "ruby",
    "s",
    "samp",
    "script",
    "section",
    "select",
    "small",
    "source",
    "span",
    "strong",
    "style",
    "sub",
    "summary",
    "sup",
    "table",
    "tbody",
    "td",
    "textarea",
    "tfoot",
    "th",
    "thead",
    "time",
    "title",
    "tr",
    "track",
    "u",
    "ul",
    "var",
    "video",
    "wbr",
    "webview",
]);
/**
 * @internal
 */
var svgElements = svgTuple([
    "animate",
    "circle",
    "clipPath",
    "defs",
    "desc",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feDropShadow",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    "filter",
    "foreignObject",
    "g",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "metadata",
    "path",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "stop",
    "svg",
    "switch",
    "symbol",
    "text",
    "textPath",
    "tspan",
    "use",
    "view",
]);
var elements = htmlElements.concat(svgElements);

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
}

/**
 * Uses the ref that is passed in, or creates a new one
 * @param external - External ref
 * @internal
 */
function useExternalRef(external) {
    // We're conditionally calling `useRef` here which is sort of naughty as hooks
    // shouldn't be called conditionally. However, Framer Motion will break if this
    // condition changes anyway. It might be possible to use an invariant here to
    // make it explicit, but I expect changing `ref` is not normal behaviour.
    var ref = !external || typeof external === "function" ? useRef(null) : external;
    useEffect(function () {
        if (external && typeof external === "function") {
            external(ref.current);
            return function () { return external(null); };
        }
    }, []);
    return ref;
}

var isFloat = function (value) {
    return !isNaN(parseFloat(value));
};
// This function is unrolled and mutative, as the number of props
// is too small to really warrent a loop and by this point config
// has been recreated many times over.
var parseDurations = function (config) {
    if (config["duration"]) {
        config["duration"] = config["duration"] * 1000;
    }
    if (config.delay) {
        config.delay = config.delay * 1000;
    }
    return config;
};
/**
 * `MotionValue` is used to track the state and velocity of motion values.
 *
 * @public
 */
var MotionValue = /** @class */ (function () {
    /**
     * @param init - The initiating value
     * @param config - Optional configuration options
     *
     * -  `transformer`: A function to transform incoming values with.
     *
     * @internal
     */
    function MotionValue(init, _a) {
        var _b = _a === void 0 ? {} : _a, transformer = _b.transformer, parent = _b.parent;
        var _this = this;
        /**
         * Duration, in milliseconds, since last updating frame.
         *
         * @internal
         */
        this.timeDelta = 0;
        /**
         * Timestamp of the last time this `MotionValue` was updated.
         *
         * @internal
         */
        this.lastUpdated = 0;
        /**
         * Tracks whether this value can output a velocity. Currently this is only true
         * if the value is numerical, but we might be able to widen the scope here and support
         * other value types.
         *
         * @internal
         */
        this.canTrackVelocity = false;
        /**
         * Notify a subscriber with the latest value.
         *
         * This is an instanced and bound function to prevent generating a new
         * function once per frame.
         *
         * @param subscriber - The subscriber to notify.
         *
         * @internal
         */
        this.notifySubscriber = function (subscriber) {
            subscriber(_this.current);
        };
        /**
         * Schedule a velocity check for the next frame.
         *
         * This is an instanced and bound function to prevent generating a new
         * function once per frame.
         *
         * @internal
         */
        this.scheduleVelocityCheck = function () { return sync.postRender(_this.velocityCheck); };
        /**
         * Updates `prev` with `current` if the value hasn't been updated this frame.
         * This ensures velocity calculations return `0`.
         *
         * This is an instanced and bound function to prevent generating a new
         * function once per frame.
         *
         * @internal
         */
        this.velocityCheck = function (_a) {
            var timestamp = _a.timestamp;
            if (timestamp !== _this.lastUpdated) {
                _this.prev = _this.current;
            }
        };
        /**
         * Updates child `MotionValue`.
         *
         * @param child - Child `MotionValue`.
         *
         * @internal
         */
        this.setChild = function (child) { return child.set(_this.current); };
        this.parent = parent;
        this.transformer = transformer;
        this.set(init, false);
        this.canTrackVelocity = isFloat(this.current);
    }
    /**
     * Creates a new `MotionValue` that's subscribed to the output of this one.
     *
     * @param config - Optional configuration options
     *
     * -  `transformer`: A function to transform incoming values with.
     *
     * @internal
     */
    MotionValue.prototype.addChild = function (config) {
        var child = new MotionValue(this.current, __assign({ parent: this }, config));
        if (!this.children)
            this.children = new Set();
        this.children.add(child);
        return child;
    };
    /**
     * Stops a `MotionValue` from being subscribed to this one.
     *
     * @param child - The subscribed `MotionValue`
     *
     * @internal
     */
    MotionValue.prototype.removeChild = function (child) {
        if (!this.children) {
            return;
        }
        this.children.delete(child);
    };
    /**
     * Subscribes a subscriber function to a subscription list.
     *
     * @param subscriptions - A `Set` of subscribers.
     * @param subscription - A subscriber function.
     */
    MotionValue.prototype.subscribeTo = function (subscriptions, subscription) {
        var _this = this;
        var updateSubscriber = function () { return subscription(_this.current); };
        subscriptions.add(updateSubscriber);
        return function () { return subscriptions.delete(updateSubscriber); };
    };
    /**
     * Adds a function that will be notified when the `MotionValue` is updated.
     *
     * @param subscriber - A function that's provided the latest value.
     * @returns A function that, when called, will cancel this subscription.
     *
     * @public
     */
    MotionValue.prototype.onChange = function (subscription) {
        if (!this.updateSubscribers)
            this.updateSubscribers = new Set();
        return this.subscribeTo(this.updateSubscribers, subscription);
    };
    /**
     * Adds a function that will be notified when the `MotionValue` requests a render.
     *
     * @param subscriber - A function that's provided the latest value.
     * @returns A function that, when called, will cancel this subscription.
     *
     * @internal
     */
    MotionValue.prototype.onRenderRequest = function (subscription) {
        if (!this.renderSubscribers)
            this.renderSubscribers = new Set();
        // Render immediately
        this.notifySubscriber(subscription);
        return this.subscribeTo(this.renderSubscribers, subscription);
    };
    /**
     * Sets the state of the `MotionValue`.
     *
     * @remarks
     *
     * ```jsx
     * const x = useMotionValue(0)
     * x.set(10)
     * ```
     *
     * @param latest - Latest value to set.
     * @param render - Whether to notify render subscribers. Defaults to `true`
     *
     * @public
     */
    MotionValue.prototype.set = function (v, render) {
        if (render === void 0) { render = true; }
        this.prev = this.current;
        this.current = this.transformer ? this.transformer(v) : v;
        if (this.updateSubscribers) {
            this.updateSubscribers.forEach(this.notifySubscriber);
        }
        if (this.children) {
            this.children.forEach(this.setChild);
        }
        if (render && this.renderSubscribers) {
            this.renderSubscribers.forEach(this.notifySubscriber);
        }
        // Update timestamp
        var _a = getFrameData(), delta = _a.delta, timestamp = _a.timestamp;
        if (this.lastUpdated !== timestamp) {
            this.timeDelta = delta;
            this.lastUpdated = timestamp;
            sync.postRender(this.scheduleVelocityCheck);
        }
    };
    /**
     * Returns the latest state of `MotionValue`
     *
     * @returns - The latest state of `MotionValue`
     *
     * @public
     */
    MotionValue.prototype.get = function () {
        return this.current;
    };
    /**
     * Returns the latest velocity of `MotionValue`
     *
     * @returns - The latest velocity of `MotionValue`. Returns `0` if the state is non-numerical.
     *
     * @public
     */
    MotionValue.prototype.getVelocity = function () {
        // This could be isFloat(this.prev) && isFloat(this.current), but that would be wasteful
        return this.canTrackVelocity
            ? // These casts could be avoided if parseFloat would be typed better
                velocityPerSecond(parseFloat(this.current) -
                    parseFloat(this.prev), this.timeDelta)
            : 0;
    };
    /**
     * Registers a new Popmotion animation to control this `MotionValue`. Only one
     * animation can drive a `MotionValue` at one time.
     *
     * @param action - The Popmotion animation or input to initialise.
     * @param config - The configuration to pass to the Popmotion animation.
     * @param transformer - An optional function to transform the output of the Popmotion animation.
     *
     * @internal
     */
    MotionValue.prototype.control = function (controller, config, transformer) {
        var _this = this;
        this.stop();
        var _a = parseDurations(config), delay$$1 = _a.delay, timeAdjustedConfig = __rest(_a, ["delay"]);
        var initialisedController = controller(timeAdjustedConfig);
        if (transformer) {
            initialisedController = initialisedController.pipe(transformer);
        }
        if (delay$$1) {
            initialisedController = chain(delay(delay$$1), initialisedController);
        }
        return new Promise(function (complete) {
            _this.controller = initialisedController.start({
                update: function (v) { return _this.set(v); },
                complete: complete,
            });
        });
    };
    /**
     * Stop the currently controlling animation.
     *
     * @public
     */
    MotionValue.prototype.stop = function () {
        if (this.controller)
            this.controller.stop();
    };
    /**
     * Destroy and clean up this `MotionValue`.
     *
     * @public
     */
    MotionValue.prototype.destroy = function () {
        this.updateSubscribers && this.updateSubscribers.clear();
        this.renderSubscribers && this.renderSubscribers.clear();
        this.parent && this.parent.removeChild(this);
        this.stop();
    };
    return MotionValue;
}());
/**
 * @internal
 */
function motionValue(init, opts) {
    return new MotionValue(init, opts);
}

// Creating a styler factory for the `onUpdate` prop allows all values
// to fire and the `onUpdate` prop will only fire once per frame
var updateStyler = createStylerFactory({
    onRead: function () { return null; },
    onRender: function (state, _a) {
        var onUpdate = _a.onUpdate;
        return onUpdate(state);
    },
});
var MotionValuesMap = /** @class */ (function () {
    function MotionValuesMap(isStatic) {
        this.hasMounted = false;
        this.values = new Map();
        this.unsubscribers = new Map();
        this.isStatic = isStatic;
    }
    MotionValuesMap.prototype.has = function (key) {
        return this.values.has(key);
    };
    MotionValuesMap.prototype.set = function (key, value) {
        this.values.set(key, value);
        if (this.hasMounted) {
            this.bindValueToStyler(key, value);
        }
    };
    MotionValuesMap.prototype.get = function (key, defaultValue) {
        var value = this.values.get(key);
        if (value === undefined && defaultValue !== undefined) {
            value = new MotionValue(defaultValue);
            this.set(key, value);
        }
        return value;
    };
    MotionValuesMap.prototype.forEach = function (callback) {
        return this.values.forEach(callback);
    };
    MotionValuesMap.prototype.bindValueToStyler = function (key, value) {
        var _this = this;
        var update = function (v) {
            // If this is a custom value, resolve it and then set the return value
            _this.styler.set(key, v);
            // If these have been changed by a custom value, add those to onUpdate
            if (_this.onUpdate) {
                _this.onUpdate.set(key, v);
            }
        };
        var unsubscribe = value.onRenderRequest(update);
        this.unsubscribers.set(key, unsubscribe);
    };
    MotionValuesMap.prototype.setOnUpdate = function (onUpdate) {
        this.onUpdate = undefined;
        if (onUpdate) {
            this.onUpdate = updateStyler({ onUpdate: onUpdate });
        }
    };
    MotionValuesMap.prototype.setTransformTemplate = function (transformTemplate) {
        if (this.transformTemplate !== transformTemplate) {
            this.transformTemplate = transformTemplate;
            this.updateTransformTemplate();
        }
    };
    MotionValuesMap.prototype.getTransformTemplate = function () {
        return this.transformTemplate;
    };
    MotionValuesMap.prototype.updateTransformTemplate = function () {
        if (this.styler) {
            this.styler.set("transform", this.transformTemplate);
        }
    };
    MotionValuesMap.prototype.mount = function (element) {
        var _this = this;
        this.hasMounted = true;
        this.styler = styler(element, {
            preparseOutput: false,
            enableHardwareAcceleration: !this.isStatic,
        });
        this.values.forEach(function (value, key) { return _this.bindValueToStyler(key, value); });
        this.updateTransformTemplate();
    };
    MotionValuesMap.prototype.unmount = function () {
        var _this = this;
        this.values.forEach(function (_value, key) {
            var unsubscribe = _this.unsubscribers.get(key);
            unsubscribe && unsubscribe();
        });
    };
    return MotionValuesMap;
}());
var useMotionValues = function (_a, isStatic) {
    var onUpdate = _a.onUpdate, transformTemplate = _a.transformTemplate;
    var motionValues = useMemo(function () { return new MotionValuesMap(isStatic); }, []);
    motionValues.setOnUpdate(onUpdate);
    motionValues.setTransformTemplate(transformTemplate);
    return motionValues;
};
/**
 * `useEffect` gets resolved bottom-up. We defer some optional functionality to child
 * components, so to ensure everything runs correctly we export the ref-binding logic
 * to a new component rather than in `useMotionValues`.
 */
var MountMotionValuesComponent = function (_a, ref) {
    var values = _a.values;
    useEffect(function () {
        invariant(ref.current instanceof Element, "No `ref` found. Ensure components created with `motion.custom` forward refs using `React.forwardRef`");
        values.mount(ref.current);
        return function () { return values.unmount(); };
    });
    return null;
};
var MountMotionValues = memo(forwardRef(MountMotionValuesComponent));

var createValueResolver = function (resolver) { return function (values) {
    var resolvedValues = {};
    values.forEach(function (value, key) { return (resolvedValues[key] = resolver(value)); });
    return resolvedValues;
}; };
var resolveCurrent = createValueResolver(function (value) { return value.get(); });

var transformOriginProps = new Set(["originX", "originY"]);
var isTransformOriginProp = function (key) { return transformOriginProps.has(key); };
var isMotionValue = function (value) {
    return value instanceof MotionValue;
};
var buildStyleAttr = function (values, styleProp, isStatic) {
    return __assign({}, styleProp, buildStyleProperty(__assign({ transform: values.getTransformTemplate() }, resolveCurrent(values)), !isStatic));
};
var useMotionStyles = function (values, styleProp, transformValues) {
    if (styleProp === void 0) { styleProp = {}; }
    var style = useRef({}).current;
    var prevMotionStyles = useRef({}).current;
    var currentStyleKeys = new Set(Object.keys(style));
    for (var key in styleProp) {
        currentStyleKeys.delete(key);
        var thisStyle = styleProp[key];
        if (isMotionValue(thisStyle)) {
            // If this is a motion value, add it to our MotionValuesMap
            values.set(key, thisStyle);
        }
        else if (isTransformProp(key) || isTransformOriginProp(key)) {
            // Or if it's a transform prop, create a motion value (or update an existing one)
            // to ensure Stylefire can reconcile all the transform values together.
            if (!values.has(key)) {
                // If it doesn't exist as a motion value, create it
                values.set(key, motionValue(thisStyle));
            }
            else {
                // Otherwise only update it if it's changed from a previous render
                if (thisStyle !== prevMotionStyles[key]) {
                    var value = values.get(key);
                    value.set(thisStyle);
                }
            }
            prevMotionStyles[key] = thisStyle;
        }
        else {
            style[key] = thisStyle;
        }
    }
    currentStyleKeys.forEach(function (key) { return delete style[key]; });
    return transformValues ? transformValues(style) : style;
};

var isKeyframesTarget = function (v) {
    return Array.isArray(v);
};

var underDampedSpring = function () { return ({
    type: "spring",
    stiffness: 500,
    damping: 25,
    restDelta: 0.5,
    restSpeed: 10,
}); };
var overDampedSpring = function (to) { return ({
    type: "spring",
    stiffness: 700,
    damping: to === 0 ? 100 : 35,
}); };
var linearTween = function () { return ({
    ease: "linear",
    duration: 0.3,
}); };
var keyframes$1 = function (values) { return ({
    type: "keyframes",
    duration: 0.8,
    values: values,
}); };
var defaultTransitions = {
    x: underDampedSpring,
    y: underDampedSpring,
    z: underDampedSpring,
    rotate: underDampedSpring,
    rotateX: underDampedSpring,
    rotateY: underDampedSpring,
    rotateZ: underDampedSpring,
    scaleX: overDampedSpring,
    scaleY: overDampedSpring,
    scale: overDampedSpring,
    opacity: linearTween,
    backgroundColor: linearTween,
    color: linearTween,
    default: overDampedSpring,
};
var getDefaultTransition = function (valueKey, to) {
    var transitionFactory;
    if (isKeyframesTarget(to)) {
        transitionFactory = keyframes$1;
    }
    else {
        transitionFactory =
            defaultTransitions[valueKey] || defaultTransitions.default;
    }
    return __assign({ to: to }, transitionFactory(to));
};

var just = function (_a) {
    var to = _a.to;
    return action(function (_a) {
        var update = _a.update, complete = _a.complete;
        update(to);
        complete();
    });
};
var transitions = { tween: tween, spring: spring, keyframes: keyframes, physics: physics, inertia: inertia, just: just };
var linear = easing.linear, easeIn = easing.easeIn, easeOut = easing.easeOut, easeInOut = easing.easeInOut, circIn = easing.circIn, circOut = easing.circOut, circInOut = easing.circInOut, backIn = easing.backIn, backOut = easing.backOut, backInOut = easing.backInOut, anticipate = easing.anticipate;
var easingLookup = {
    linear: linear,
    easeIn: easeIn,
    easeOut: easeOut,
    easeInOut: easeInOut,
    circIn: circIn,
    circOut: circOut,
    circInOut: circInOut,
    backIn: backIn,
    backOut: backOut,
    backInOut: backInOut,
    anticipate: anticipate,
};
var easingDefinitionToFunction = function (definition) {
    if (Array.isArray(definition)) {
        // If cubic bezier definition, create bezier curve
        invariant(definition.length === 4, "Cubic bezier arrays must contain four numerical values.");
        var x1 = definition[0], y1 = definition[1], x2 = definition[2], y2 = definition[3];
        return easing.cubicBezier(x1, y1, x2, y2);
    }
    else if (typeof definition === "string") {
        // Else lookup from table
        invariant(easingLookup[definition] !== undefined, "Invalid easing type '" + definition + "'");
        return easingLookup[definition];
    }
    return definition;
};
var transitionOptionParser = {
    tween: function (opts) {
        if (opts.ease) {
            opts.ease = easingDefinitionToFunction(opts.ease);
        }
        return opts;
    },
    keyframes: function (_a) {
        var from = _a.from, to = _a.to, velocity = _a.velocity, opts = __rest(_a, ["from", "to", "velocity"]);
        if (opts.values && opts.values[0] === null) {
            var values = opts.values.slice();
            values[0] = from;
            opts.values = values;
        }
        if (opts.ease) {
            opts.ease = easingDefinitionToFunction(opts.ease);
        }
        if (opts.easings) {
            opts.easings = opts.easings.map(easingDefinitionToFunction);
        }
        return opts;
    },
};
var isTransitionDefined = function (_a) {
    var when = _a.when, delay$$1 = _a.delay, delayChildren = _a.delayChildren, staggerChildren = _a.staggerChildren, staggerDirection = _a.staggerDirection, transition = __rest(_a, ["when", "delay", "delayChildren", "staggerChildren", "staggerDirection"]);
    return Object.keys(transition).length;
};
var getTransitionForValue = function (key, to, transitionDefinition) {
    var delay$$1 = transitionDefinition ? transitionDefinition.delay : 0;
    // If no object, return default transition
    // A better way to handle this would be to deconstruct out all the shared Orchestration props
    // and see if there's any props remaining
    if (transitionDefinition === undefined ||
        !isTransitionDefined(transitionDefinition)) {
        return __assign({ delay: delay$$1 }, getDefaultTransition(key, to));
    }
    var valueTransitionDefinition = transitionDefinition[key] ||
        transitionDefinition.default ||
        transitionDefinition;
    if (valueTransitionDefinition.type === false) {
        return {
            type: "just",
            delay: delay$$1,
            to: isKeyframesTarget(to)
                ? to[to.length - 1]
                : to,
        };
    }
    else if (isKeyframesTarget(to)) {
        return __assign({ values: to, duration: 0.8, delay: delay$$1, ease: "linear" }, valueTransitionDefinition, { 
            // This animation must be keyframes if we're animating through an array
            type: "keyframes" });
    }
    else {
        return __assign({ type: "tween", to: to,
            delay: delay$$1 }, valueTransitionDefinition);
    }
};
var preprocessOptions = function (type, opts) {
    return transitionOptionParser[type]
        ? transitionOptionParser[type](opts)
        : opts;
};
var getTransition = function (value, valueKey, to, transition) {
    var _a = getTransitionForValue(valueKey, to, transition), _b = _a.type, type = _b === void 0 ? "tween" : _b, transitionDefinition = __rest(_a, ["type"]);
    var actionFactory = transitions[type];
    var opts = preprocessOptions(type, __assign({ from: value.get(), velocity: value.getVelocity() }, transitionDefinition));
    return [actionFactory, opts];
};

var auto = {
    test: function (v) { return v === "auto"; },
    parse: function (v) { return v; },
};
var valueTypeTests = [number, px, percent, degrees, vw, vh, auto];
var testValueType = function (v) { return function (type) { return type.test(v); }; };
var getValueType = function (v) { return valueTypeTests.find(testValueType(v)); };

var positionalKeys = new Set([
    "width",
    "height",
    "top",
    "left",
    "right",
    "bottom",
    "x",
    "y",
]);
var isPositionalKey = function (key) { return positionalKeys.has(key); };
var hasPositionalKey = function (target) {
    return Object.keys(target).some(isPositionalKey);
};
var setAndResetVelocity = function (value, to) {
    // Looks odd but setting it twice doesn't render, it'll just
    // set both prev and current to the latest value
    value.set(to, false);
    value.set(to);
};
var isNumOrPxType = function (v) {
    return v === number || v === px;
};
var BoundingBoxDimension;
(function (BoundingBoxDimension) {
    BoundingBoxDimension["width"] = "width";
    BoundingBoxDimension["height"] = "height";
    BoundingBoxDimension["left"] = "left";
    BoundingBoxDimension["right"] = "right";
    BoundingBoxDimension["top"] = "top";
    BoundingBoxDimension["bottom"] = "bottom";
})(BoundingBoxDimension || (BoundingBoxDimension = {}));
var getPosFromMatrix = function (matrix, pos) {
    return parseFloat(matrix.split(", ")[pos]);
};
var getTranslateFromMatrix = function (pos2, pos3) { return function (_bbox, _a) {
    var transform = _a.transform;
    if (transform === "none" || !transform)
        return 0;
    var matrix3d = transform.match(/^matrix3d\((.+)\)$/);
    if (matrix3d) {
        return getPosFromMatrix(matrix3d[1], pos3);
    }
    else {
        var matrix = transform.match(/^matrix\((.+)\)$/);
        return getPosFromMatrix(matrix[1], pos2);
    }
}; };
var positionalValues = {
    // Dimensions
    width: function (_a) {
        var width = _a.width;
        return width;
    },
    height: function (_a) {
        var height = _a.height;
        return height;
    },
    top: function (_bbox, _a) {
        var top = _a.top;
        return parseFloat(top);
    },
    left: function (_bbox, _a) {
        var left = _a.left;
        return parseFloat(left);
    },
    bottom: function (_a, _b) {
        var height = _a.height;
        var top = _b.top;
        return parseFloat(top) + height;
    },
    right: function (_a, _b) {
        var width = _a.width;
        var left = _b.left;
        return parseFloat(left) + width;
    },
    // Transform
    x: getTranslateFromMatrix(4, 13),
    y: getTranslateFromMatrix(5, 14),
};
var convertChangedValueTypes = function (target, values, ref, changedKeys) {
    var element = ref.current;
    var elementStyler = styler(element);
    var originBbox = element.getBoundingClientRect();
    var elementComputedStyle = getComputedStyle(element);
    var top = elementComputedStyle.top, left = elementComputedStyle.left, bottom = elementComputedStyle.bottom, right = elementComputedStyle.right, transform = elementComputedStyle.transform;
    var originComputedStyle = { top: top, left: left, bottom: bottom, right: right, transform: transform };
    // Apply the latest values (as set in checkAndConvertChangedValueTypes)
    elementStyler.render();
    var targetBbox = element.getBoundingClientRect();
    changedKeys.forEach(function (key) {
        // Restore styles to their **calculated computed style**, not their actual
        // originally set style. This allows us to animate between equivalent pixel units.
        var value = values.get(key);
        setAndResetVelocity(value, positionalValues[key](originBbox, originComputedStyle));
        target[key] = positionalValues[key](targetBbox, elementComputedStyle);
    });
    // Reapply original values
    elementStyler.render();
    return target;
};
var checkAndConvertChangedValueTypes = function (values, ref, target, transitionEnd) {
    if (transitionEnd === void 0) { transitionEnd = {}; }
    target = __assign({}, target);
    transitionEnd = __assign({}, transitionEnd);
    var targetPositionalKeys = Object.keys(target).filter(isPositionalKey);
    var changedValueTypeKeys = targetPositionalKeys.reduce(function (acc, key) {
        var value = values.get(key);
        if (!value)
            return acc;
        var from = value.get();
        var to = target[key];
        var fromType = getValueType(from);
        var toType;
        // TODO: The current implementation of this basically throws an error
        // if you try and do value conversion via keyframes. There's probably
        // a way of doing this but the performance implications would need greater scrutiny,
        // as it'd be doing multiple resize-remeasure operations.
        if (isKeyframesTarget(to)) {
            var numKeyframes = to.length;
            for (var i = to[0] === null ? 1 : 0; i < numKeyframes; i++) {
                if (!toType) {
                    toType = getValueType(to[i]);
                    invariant(toType === fromType ||
                        (isNumOrPxType(fromType) &&
                            isNumOrPxType(toType)), "Keyframes must be of the same dimension as the current value");
                }
                else {
                    invariant(getValueType(to[i]) === toType, "All keyframes must be of the same type");
                }
            }
        }
        else {
            toType = getValueType(to);
        }
        if (fromType !== toType) {
            // If they're both just number or px, convert them both to numbers rather than
            // relying on resize/remeasure to convert (which is wasteful in this situation)
            if (isNumOrPxType(fromType) && isNumOrPxType(toType)) {
                var current = value.get();
                if (typeof current === "string") {
                    value.set(parseFloat(current));
                }
                if (typeof to === "string") {
                    target[key] = parseFloat(to);
                }
                else if (Array.isArray(to) && toType === px) {
                    target[key] = to.map(parseFloat);
                }
            }
            else {
                acc.push(key);
                transitionEnd[key] =
                    transitionEnd[key] !== undefined
                        ? transitionEnd[key]
                        : target[key];
                setAndResetVelocity(value, to);
            }
        }
        return acc;
    }, []);
    return changedValueTypeKeys.length
        ? {
            target: convertChangedValueTypes(target, values, ref, changedValueTypeKeys),
            transitionEnd: transitionEnd,
        }
        : { target: target, transitionEnd: transitionEnd };
};
/**
 * Convert value types for x/y/width/height/top/left/bottom/right
 *
 * Allows animation between `'auto'` -> `'100%'` or `0` -> `'calc(50% - 10vw)'`
 *
 * @param values
 * @param ref
 * @param target
 * @param transitionEnd
 * @internal
 */
function unitConversion(values, ref, target, transitionEnd) {
    return hasPositionalKey(target)
        ? checkAndConvertChangedValueTypes(values, ref, target, transitionEnd)
        : { target: target, transitionEnd: transitionEnd };
}

var isCustomValue = function (v) {
    return Boolean(v && typeof v === "object" && v.mix && v.toValue);
};
var resolveFinalValueInKeyframes = function (v) {
    // TODO maybe throw if v.length - 1 is placeholder token?
    return isKeyframesTarget(v) ? v[v.length - 1] || 0 : v;
};

var getCurrent = function (values) {
    var current = {};
    values.forEach(function (value, key) { return (current[key] = value.get()); });
    return current;
};
var getVelocity = function (values) {
    var velocity = {};
    values.forEach(function (value, key) { return (velocity[key] = value.getVelocity()); });
    return velocity;
};
var isAnimatable = function (value) {
    if (typeof value === "number" || Array.isArray(value)) {
        return true;
    }
    else if (typeof value === "string" &&
        complex.test(value) &&
        !value.startsWith("url(")) {
        return true;
    }
    return false;
};
var isTargetResolver = function (p) {
    return typeof p === "function";
};
var isVariantLabels = function (v) { return Array.isArray(v); };
var isNumericalString = function (v) { return /^\d*\.?\d+$/.test(v); };
var ComponentAnimationControls = /** @class */ (function () {
    function ComponentAnimationControls(values, ref) {
        var _this = this;
        this.variants = {};
        this.baseTarget = {};
        this.overrides = [];
        this.activeOverrides = new Set();
        this.resolvedOverrides = [];
        this.isAnimating = new Set();
        this.values = values;
        this.ref = ref;
        this.values.forEach(function (value, key) { return (_this.baseTarget[key] = value.get()); });
    }
    ComponentAnimationControls.prototype.setProps = function (props) {
        this.props = props;
    };
    ComponentAnimationControls.prototype.setVariants = function (variants) {
        if (variants)
            this.variants = variants;
    };
    ComponentAnimationControls.prototype.setDefaultTransition = function (transition) {
        if (transition)
            this.defaultTransition = transition;
    };
    ComponentAnimationControls.prototype.setValues = function (target, isActive) {
        var _this = this;
        if (isActive === void 0) { isActive = new Set(); }
        target = this.transformValues(target);
        return Object.keys(target).forEach(function (key) {
            if (isActive.has(key))
                return;
            isActive.add(key);
            var targetValue = resolveFinalValueInKeyframes(target[key]);
            if (_this.values.has(key)) {
                var value = _this.values.get(key);
                value && value.set(targetValue);
            }
            else {
                _this.values.set(key, motionValue(targetValue));
            }
            _this.baseTarget[key] = targetValue;
        });
    };
    ComponentAnimationControls.prototype.transformValues = function (values) {
        var transformValues = this.props.transformValues;
        return transformValues ? transformValues(values) : values;
    };
    ComponentAnimationControls.prototype.checkForNewValues = function (target) {
        var _this = this;
        var newValueKeys = Object.keys(target).filter(function (key) { return !_this.values.has(key); });
        if (!newValueKeys.length)
            return;
        // Might live better in `MotionValuesMap`
        var domStyler = styler(this.ref.current);
        newValueKeys.forEach(function (key) {
            var domValue = domStyler.get(key) || 0;
            var value = typeof domValue === "string" && isNumericalString(domValue)
                ? parseFloat(domValue)
                : domValue;
            _this.values.set(key, motionValue(value));
            _this.baseTarget[key] = value;
        });
    };
    ComponentAnimationControls.prototype.resolveVariant = function (variant) {
        if (!variant) {
            return {
                target: undefined,
                transition: undefined,
                transitionEnd: undefined,
            };
        }
        if (isTargetResolver(variant)) {
            // resolve current and velocity
            variant = variant(this.props, getCurrent(this.values), getVelocity(this.values));
        }
        var _a = variant.transition, transition = _a === void 0 ? this.defaultTransition : _a, transitionEnd = variant.transitionEnd, target = __rest(variant, ["transition", "transitionEnd"]);
        return { transition: transition, transitionEnd: transitionEnd, target: target };
    };
    ComponentAnimationControls.prototype.getHighestPriority = function () {
        if (!this.activeOverrides.size)
            return 0;
        return Math.max.apply(Math, Array.from(this.activeOverrides));
    };
    ComponentAnimationControls.prototype.setOverride = function (definition, overrideIndex) {
        this.overrides[overrideIndex] = definition;
    };
    ComponentAnimationControls.prototype.startOverride = function (overrideIndex) {
        var override = this.overrides[overrideIndex];
        if (override) {
            return this.start(override, { priority: overrideIndex });
        }
    };
    ComponentAnimationControls.prototype.clearOverride = function (overrideIndex) {
        var override = this.overrides[overrideIndex];
        if (!override)
            return;
        this.activeOverrides.delete(overrideIndex);
        var highest = this.getHighestPriority();
        this.resetIsAnimating();
        if (highest) {
            var highestOverride = this.overrides[highest];
            highestOverride && this.startOverride(highest);
        }
        // Figure out which remaining values were affected by the override and animate those
        var overrideTarget = this.resolvedOverrides[overrideIndex];
        if (!overrideTarget)
            return;
        var remainingValues = {};
        for (var key in this.baseTarget) {
            if (overrideTarget[key] !== undefined) {
                remainingValues[key] = this.baseTarget[key];
            }
        }
        this.animate(remainingValues);
    };
    ComponentAnimationControls.prototype.apply = function (definition) {
        if (Array.isArray(definition)) {
            return this.applyVariantLabels(definition);
        }
        else if (typeof definition === "string") {
            return this.applyVariantLabels([definition]);
        }
        else {
            this.setValues(definition);
        }
    };
    ComponentAnimationControls.prototype.applyVariantLabels = function (variantLabelList) {
        var _this = this;
        var isSetting = new Set();
        var reversedList = variantLabelList.slice().reverse();
        reversedList.forEach(function (key) {
            var _a = _this.resolveVariant(_this.variants[key]), target = _a.target, transitionEnd = _a.transitionEnd;
            if (transitionEnd) {
                _this.setValues(transitionEnd, isSetting);
            }
            if (target) {
                _this.setValues(target, isSetting);
            }
        });
    };
    ComponentAnimationControls.prototype.start = function (definition, opts) {
        var _this = this;
        if (opts === void 0) { opts = {}; }
        if (opts.priority) {
            this.activeOverrides.add(opts.priority);
        }
        this.resetIsAnimating(opts.priority);
        var animation;
        if (isVariantLabels(definition)) {
            animation = this.animateVariantLabels(definition, opts);
        }
        else if (typeof definition === "string") {
            animation = this.animateVariant(definition, opts);
        }
        else {
            animation = this.animate(definition, opts);
        }
        return animation.then(function () {
            var onAnimationComplete = _this.props.onAnimationComplete;
            onAnimationComplete && onAnimationComplete();
        });
    };
    ComponentAnimationControls.prototype.isHighestPriority = function (priority) {
        var numOverrides = this.overrides.length;
        for (var i = priority + 1; i < numOverrides; i++) {
            if (this.overrides[i] !== undefined) {
                return false;
            }
        }
        return true;
    };
    ComponentAnimationControls.prototype.animate = function (animationDefinition, _a) {
        var _this = this;
        var _b = _a === void 0 ? {} : _a, _c = _b.delay, delay$$1 = _c === void 0 ? 0 : _c, _d = _b.priority, priority = _d === void 0 ? 0 : _d, transitionOverride = _b.transitionOverride;
        var _e = this.resolveVariant(animationDefinition), target = _e.target, transition = _e.transition, transitionEnd = _e.transitionEnd;
        if (transitionOverride) {
            transition = transitionOverride;
        }
        if (!target)
            return Promise.resolve();
        target = this.transformValues(target);
        if (transitionEnd) {
            transitionEnd = this.transformValues(transitionEnd);
        }
        if (priority) {
            this.resolvedOverrides[priority] = target;
        }
        this.checkForNewValues(target);
        var converted = unitConversion(this.values, this.ref, target, transitionEnd);
        target = converted.target;
        transitionEnd = converted.transitionEnd;
        // TODO: This might be redundant (see `resolveVariant`)
        if (!transition && this.defaultTransition) {
            transition = this.defaultTransition;
        }
        var animations = Object.keys(target).reduce(function (acc, key) {
            var value = _this.values.get(key);
            if (!value || !target || target[key] === undefined)
                return acc;
            var valueTarget = target[key];
            if (!priority) {
                _this.baseTarget[key] = resolveFinalValueInKeyframes(valueTarget);
            }
            if (_this.isAnimating.has(key))
                return acc;
            if (isAnimatable(valueTarget)) {
                var _a = getTransition(value, key, valueTarget, __assign({ delay: delay$$1 }, transition)), action$$1 = _a[0], options = _a[1];
                acc.push(value.control(action$$1, options));
            }
            else {
                value.set(valueTarget);
            }
            _this.isAnimating.add(key);
            return acc;
        }, []);
        return Promise.all(animations).then(function () {
            if (!transitionEnd)
                return;
            _this.setValues(transitionEnd);
        });
    };
    ComponentAnimationControls.prototype.animateVariantLabels = function (variantLabels, opts) {
        var _this = this;
        var animations = variantLabels.slice().reverse()
            .map(function (label) { return _this.animateVariant(label, opts); });
        return Promise.all(animations);
    };
    ComponentAnimationControls.prototype.animateVariant = function (variantLabel, opts) {
        var _this = this;
        var when = false;
        var delayChildren = 0;
        var staggerChildren = 0;
        var staggerDirection = 1;
        var priority = (opts && opts.priority) || 0;
        var variant = this.variants[variantLabel];
        var getAnimations = variant
            ? function () { return _this.animate(variant, opts); }
            : function () { return Promise.resolve(); };
        var getChildrenAnimations = this.children
            ? function () {
                return _this.animateChildren(variantLabel, delayChildren, staggerChildren, staggerDirection, priority);
            }
            : function () { return Promise.resolve(); };
        if (variant && this.children) {
            var transition = this.resolveVariant(variant).transition;
            if (transition) {
                when = transition.when || when;
                delayChildren = transition.delayChildren || delayChildren;
                staggerChildren = transition.staggerChildren || staggerChildren;
                staggerDirection =
                    transition.staggerDirection || staggerDirection;
            }
        }
        if (when) {
            var _a = when === "beforeChildren"
                ? [getAnimations, getChildrenAnimations]
                : [getChildrenAnimations, getAnimations], first = _a[0], last = _a[1];
            return first().then(last);
        }
        else {
            return Promise.all([getAnimations(), getChildrenAnimations()]);
        }
    };
    ComponentAnimationControls.prototype.animateChildren = function (variantLabel, delayChildren, staggerChildren, staggerDirection, priority) {
        if (delayChildren === void 0) { delayChildren = 0; }
        if (staggerChildren === void 0) { staggerChildren = 0; }
        if (staggerDirection === void 0) { staggerDirection = 1; }
        if (priority === void 0) { priority = 0; }
        if (!this.children) {
            return Promise.resolve();
        }
        var animations = [];
        var maxStaggerDuration = (this.children.size - 1) * staggerChildren;
        var generateStaggerDuration = staggerDirection === 1
            ? function (i) { return i * staggerChildren; }
            : function (i) { return maxStaggerDuration - i * staggerChildren; };
        Array.from(this.children).forEach(function (childControls, i) {
            var animation = childControls.animateVariant(variantLabel, {
                priority: priority,
                delay: delayChildren + generateStaggerDuration(i),
            });
            animations.push(animation);
        });
        return Promise.all(animations);
    };
    ComponentAnimationControls.prototype.checkOverrideIsAnimating = function (priority) {
        var numOverrides = this.overrides.length;
        for (var i = priority + 1; i < numOverrides; i++) {
            var resolvedOverride = this.resolvedOverrides[i];
            if (resolvedOverride) {
                for (var key in resolvedOverride) {
                    this.isAnimating.add(key);
                }
            }
        }
    };
    ComponentAnimationControls.prototype.resetIsAnimating = function (priority) {
        if (priority === void 0) { priority = 0; }
        this.isAnimating.clear();
        // If this isn't the highest priority gesture, block the animation
        // of anything that's currently being animated
        if (priority < this.getHighestPriority()) {
            this.checkOverrideIsAnimating(priority);
        }
        if (this.children) {
            this.children.forEach(function (child) { return child.resetIsAnimating(priority); });
        }
    };
    ComponentAnimationControls.prototype.stop = function () {
        this.values.forEach(function (value) { return value.stop(); });
    };
    ComponentAnimationControls.prototype.addChild = function (controls) {
        if (!this.children) {
            this.children = new Set();
        }
        this.children.add(controls);
    };
    ComponentAnimationControls.prototype.removeChild = function (controls) {
        if (!this.children) {
            return;
        }
        this.children.delete(controls);
    };
    ComponentAnimationControls.prototype.resetChildren = function () {
        if (this.children)
            this.children.clear();
    };
    return ComponentAnimationControls;
}());

/**
 * Use a callback a maximum number of times
 * @param callback - Callback to run
 * @param times - Maximum number of times to run the callback. Defaults to `1`
 *
 * @public
 */
function useMaxTimes(callback, times) {
    if (times === void 0) { times = 1; }
    var count = useRef(0);
    if (count.current++ < times)
        callback();
}

/**
 * @internal
 */
var MotionContext = createContext({
    static: false,
});
var isTarget = function (v) {
    return v !== undefined && typeof v !== "string" && !Array.isArray(v);
};
var useMotionContext = function (parentContext, controls, isStatic, initial) {
    if (isStatic === void 0) { isStatic = false; }
    var context = useMemo(function () { return ({
        controls: controls,
        initial: initial && !isTarget(initial) ? initial : parentContext.initial,
    }); }, []);
    context.static = isStatic;
    // Set initial state. If this is a static component (ie in Framer canvas), respond to updates
    // in `initial`
    useMaxTimes(function () {
        var initialToApply = initial || parentContext.initial;
        initialToApply && controls.apply(initialToApply);
    }, isStatic ? Infinity : 1);
    return context;
};

/**
 * Creates an imperative set of controls to trigger animations.
 *
 * This allows a consolidated, uniform API for animations, to be triggered by other APIs like the `animate` prop, or the gesture handlers.
 *
 * @param values
 * @param props
 * @param ref
 * @param inherit
 *
 * @internal
 */
function useComponentAnimationControls(values, props, ref, inherit) {
    var variants = props.variants, transition = props.transition;
    var parentControls = useContext(MotionContext).controls;
    var controls = useMemo(function () { return new ComponentAnimationControls(values, ref); }, []);
    // Reset and resubscribe children every render to ensure stagger order is correct
    controls.resetChildren();
    if (inherit && parentControls) {
        parentControls.addChild(controls);
    }
    useEffect(function () { return function () { return parentControls && parentControls.removeChild(controls); }; }, []);
    controls.setProps(props);
    controls.setVariants(variants);
    controls.setDefaultTransition(transition);
    return controls;
}

/**
 * Control animations on one or more components.
 *
 * @public
 */
var AnimationControls = /** @class */ (function () {
    function AnimationControls() {
        /**
         * Track whether the host component has mounted.
         *
         * @internal
         */
        this.hasMounted = false;
        /**
         * Pending animations that are started before a component is mounted.
         *
         * @internal
         */
        this.pendingAnimations = [];
        /**
         * A collection of linked component animation controls.
         *
         * @internal
         */
        this.componentControls = new Set();
        /**
         * A map of variants that can be later referenced via `start(variantLabel)`
         *
         * @internal
         */
        this.variants = {};
    }
    /**
     * Set variants on this and all child components.
     *
     * @param variants
     *
     * @internal
     */
    AnimationControls.prototype.setVariants = function (variants) {
        this.variants = variants;
        this.componentControls.forEach(function (controls) {
            return controls.setVariants(variants);
        });
    };
    /**
     * Set a default transition on this and all child components
     *
     * @param transition
     *
     * @internal
     */
    AnimationControls.prototype.setDefaultTransition = function (transition) {
        this.defaultTransition = transition;
        this.componentControls.forEach(function (controls) {
            return controls.setDefaultTransition(transition);
        });
    };
    /**
     * Subscribes a component's animation controls to this.
     *
     * @param controls
     * @returns An unsubscribe function.
     *
     * @internal
     */
    AnimationControls.prototype.subscribe = function (controls) {
        var _this = this;
        this.componentControls.add(controls);
        if (this.variants)
            controls.setVariants(this.variants);
        if (this.defaultTransition)
            controls.setDefaultTransition(this.defaultTransition);
        return function () { return _this.componentControls.delete(controls); };
    };
    /**
     * Starts an animation on all linked components.
     *
     * @remarks
     *
     * ```jsx
     * animation.start("variantLabel")
     * animation.start({
     *   x: 0,
     *   transition: { duration: 1 }
     * })
     * ```
     *
     * @param definition - Properties or variant label to animate to
     * @param transition - Optional `transtion` to apply to a variant
     * @returns - A `Promise` that resolves when all animations have completed.
     *
     * @public
     */
    AnimationControls.prototype.start = function (definition, transitionOverride) {
        var _this = this;
        if (this.hasMounted) {
            var animations_1 = [];
            this.componentControls.forEach(function (controls) {
                var animation = controls.start(definition, {
                    transitionOverride: transitionOverride,
                });
                animations_1.push(animation);
            });
            return Promise.all(animations_1);
        }
        else {
            return new Promise(function (resolve) {
                _this.pendingAnimations.push({
                    animation: [definition, transitionOverride],
                    resolve: resolve,
                });
            });
        }
    };
    /**
     * Stops animations on all linked components.
     *
     * @public
     */
    AnimationControls.prototype.stop = function () {
        this.componentControls.forEach(function (controls) { return controls.stop(); });
    };
    /**
     * Initialises the animation controls.
     *
     * @internal
     */
    AnimationControls.prototype.mount = function () {
        var _this = this;
        this.hasMounted = true;
        this.pendingAnimations.forEach(function (_a) {
            var animation = _a.animation, resolve = _a.resolve;
            return _this.start.apply(_this, animation).then(resolve);
        });
    };
    /**
     * Stops all child animations when the host component unmounts.
     *
     * @internal
     */
    AnimationControls.prototype.unmount = function () {
        this.hasMounted = false;
        this.stop();
    };
    return AnimationControls;
}());
/**
 * @internal
 */
var animationControls = function () { return new AnimationControls(); };

/**
 * `useAnimationGroupSubscription` allows a component to subscribe to an
 * externally-created `AnimationControls`, created by the `useAnimation` hook.
 *
 * @param animation
 * @param controls
 *
 * @internal
 */
function useAnimationGroupSubscription(animation, controls) {
    var unsubscribe = useMemo(function () { return animation.subscribe(controls); }, [
        animation,
    ]);
    useEffect(function () { return function () { return unsubscribe && unsubscribe(); }; }, [unsubscribe]);
}

var labelsToArray = function (label) {
    if (!label) {
        return [];
    }
    if (Array.isArray(label)) {
        return label;
    }
    return [label];
};
var resolveVariantLabels = function (variant) {
    var unresolvedVariant = variant instanceof MotionValue ? variant.get() : variant;
    return Array.from(new Set(labelsToArray(unresolvedVariant)));
};
/**
 * Hooks in React sometimes accept a dependency array as their final argument. (ie useEffect/useMemo)
 * When values in this array change, React re-runs the dependency. However if the array
 * contains a variable number of items, React throws an error.
 */
var asDependencyList = function (list) { return [
    list.join(","),
]; };

var hasVariantChanged = function (oldVariant, newVariant) {
    return oldVariant.join(",") !== newVariant.join(",");
};
/**
 * Handle variants and the `animate` prop when its set as variant labels.
 *
 * @param targetVariant
 * @param inherit
 * @param controls
 * @param initialVariant
 *
 * @internal
 */
function useVariants(targetVariant, inherit, controls, initialVariant) {
    var variantList = resolveVariantLabels(targetVariant);
    var hasMounted = useRef(false);
    // Fire animations when poses change
    useEffect(function () {
        // TODO: This logic might mean we don't need to load this hook at all
        if (inherit)
            return;
        if (hasMounted.current ||
            hasVariantChanged(resolveVariantLabels(initialVariant), variantList)) {
            controls.start(variantList);
        }
        hasMounted.current = true;
    }, asDependencyList(variantList));
}

function isEventTarget(target) {
    return (typeof target.addEventListener === "function" &&
        typeof target.removeEventListener === "function");
}
var eventListener = function (target, name, handler, options) {
    var startListening = function () {
        if (!target) {
            return;
        }
        target.addEventListener(name, handler, options);
    };
    var stopListening = function () {
        if (!target) {
            return;
        }
        target.removeEventListener(name, handler, options);
    };
    return [startListening, stopListening];
};
var useEvent = function (type, target, handler, options) {
    var result = undefined;
    if ((!target || isEventTarget(target)) && handler) {
        result = eventListener(target, type, handler, options);
    }
    useEffect(function () {
        var ref = target;
        if (!handler || !target || isEventTarget(target) || !ref.current) {
            return;
        }
        var _a = eventListener(ref.current, type, handler, options), start = _a[0], stop = _a[1];
        start();
        return stop;
    }, [type, target, handler, options]);
    return result;
};

var extractEventInfo = function (_a) {
    var _b = _a.pageX, pageX = _b === void 0 ? 0 : _b, _c = _a.pageY, pageY = _c === void 0 ? 0 : _c;
    return {
        point: { x: pageX, y: pageY },
    };
};
var wrapHandler = function (handler) {
    if (!handler) {
        return undefined;
    }
    var listener = function (event, info) {
        if (!info) {
            info = extractEventInfo(event);
        }
        handler(event, info);
    };
    return listener;
};

var mockWindow = {
    addEventListener: function () { },
    removeEventListener: function () { },
    dispatchEvent: function () { return false; },
    onpointerdown: false,
    onpointermove: false,
    onpointerup: false,
    ontouchstart: false,
    ontouchmove: false,
    ontouchend: false,
    onmousedown: false,
    onmousemove: false,
    onmouseup: false,
    devicePixelRatio: 1,
    scrollX: 0,
    scrollY: 0,
    location: {
        href: "",
    },
    setTimeout: function () { return 0; },
    clearTimeout: function () { },
    setInterval: function () { return 0; },
    clearInterval: function () { },
    webkitConvertPointFromPageToNode: function (_, pt) { return pt; },
};
/**
 * Creates a server-safe reference to `window`, returning a mock if none is available.
 *
 * @internal
 */
var safeWindow = typeof window === "undefined" ? mockWindow : window;

var noop = [function () { }, function () { }];
var mergeUseEventResults = function () {
    var values = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        values[_i] = arguments[_i];
    }
    if (values.every(function (v) { return v === undefined; })) {
        return noop;
    }
    var start = function () {
        for (var _i = 0, values_1 = values; _i < values_1.length; _i++) {
            var value = values_1[_i];
            if (value) {
                value[0]();
            }
        }
    };
    var stop = function () {
        for (var _i = 0, values_2 = values; _i < values_2.length; _i++) {
            var value = values_2[_i];
            if (value) {
                value[1]();
            }
        }
    };
    return [start, stop];
};
var useMouseEvents = function (_a, ref, options) {
    var onMouseDown = _a.onMouseDown, onMouseMove = _a.onMouseMove, onMouseUp = _a.onMouseUp, onMouseOver = _a.onMouseOver, onMouseOut = _a.onMouseOut, onMouseEnter = _a.onMouseEnter, onMouseLeave = _a.onMouseLeave;
    var down = useEvent("mousedown", ref, wrapHandler(onMouseDown), options);
    var move = useEvent("mousemove", ref, wrapHandler(onMouseMove), options);
    var up = useEvent("mouseup", ref, wrapHandler(onMouseUp), options);
    var over = useEvent("mouseover", ref, wrapHandler(onMouseOver), options);
    var out = useEvent("mouseout", ref, wrapHandler(onMouseOut), options);
    var enter = useEvent("mouseenter", ref, wrapHandler(onMouseEnter), options);
    var leave = useEvent("mouseleave", ref, wrapHandler(onMouseLeave), options);
    return mergeUseEventResults(down, move, up, over, out, enter, leave);
};
var useTouchEvents = function (_a, ref, options) {
    var onTouchStart = _a.onTouchStart, onTouchMove = _a.onTouchMove, onTouchEnd = _a.onTouchEnd, onTouchCancel = _a.onTouchCancel;
    var down = useEvent("touchstart", ref, wrapHandler(onTouchStart), options);
    var move = useEvent("touchmove", ref, wrapHandler(onTouchMove), options);
    var up = useEvent("touchend", ref, wrapHandler(onTouchEnd), options);
    var cancel = useEvent("touchcancel", ref, wrapHandler(onTouchCancel), options);
    return mergeUseEventResults(down, move, up, cancel);
};
var useNativePointerEvents = function (_a, ref, options) {
    var onPointerDown = _a.onPointerDown, onPointerMove = _a.onPointerMove, onPointerUp = _a.onPointerUp, onPointerCancel = _a.onPointerCancel, onPointerOver = _a.onPointerOver, onPointerOut = _a.onPointerOut, onPointerEnter = _a.onPointerEnter, onPointerLeave = _a.onPointerLeave;
    var down = useEvent("pointerdown", ref, wrapHandler(onPointerDown), options);
    var move = useEvent("pointermove", ref, wrapHandler(onPointerMove), options);
    var up = useEvent("pointerup", ref, wrapHandler(onPointerUp), options);
    var cancel = useEvent("pointercancel", ref, wrapHandler(onPointerCancel), options);
    var over = useEvent("pointerover", ref, wrapHandler(onPointerOver), options);
    var out = useEvent("pointerout", ref, wrapHandler(onPointerOut), options);
    var enter = useEvent("pointerenter", ref, wrapHandler(onPointerEnter), options);
    var leave = useEvent("pointerleave", ref, wrapHandler(onPointerLeave), options);
    return mergeUseEventResults(down, move, up, cancel, over, out, enter, leave);
};
var supportsPointerEvents = function () {
    return safeWindow.onpointerdown === null &&
        safeWindow.onpointermove === null &&
        safeWindow.onpointerup === null;
};
var supportsTouchEvents = function () {
    return safeWindow.ontouchstart === null &&
        safeWindow.ontouchmove === null &&
        safeWindow.ontouchend === null;
};
var supportsMouseEvents = function () {
    return safeWindow.onmousedown === null &&
        safeWindow.onmousemove === null &&
        safeWindow.onmouseup === null;
};
var usePointerEvents = function (eventHandlers, ref, options) {
    var onPointerDown = eventHandlers.onPointerDown, onPointerMove = eventHandlers.onPointerMove, onPointerUp = eventHandlers.onPointerUp, onPointerCancel = eventHandlers.onPointerCancel, onPointerOver = eventHandlers.onPointerOver, onPointerOut = eventHandlers.onPointerOut, onPointerEnter = eventHandlers.onPointerEnter, onPointerLeave = eventHandlers.onPointerLeave;
    var mouseEvents = {}, touchEvents = {}, pointerEvents = {};
    if (supportsPointerEvents()) {
        pointerEvents = eventHandlers;
    }
    else if (supportsTouchEvents()) {
        touchEvents = {
            onTouchStart: onPointerDown,
            onTouchMove: onPointerMove,
            onTouchEnd: onPointerUp,
            onTouchCancel: onPointerCancel,
        };
    }
    else if (supportsMouseEvents()) {
        mouseEvents = {
            onMouseDown: onPointerDown,
            onMouseMove: onPointerMove,
            onMouseUp: onPointerUp,
            onMouseOver: onPointerOver,
            onMouseOut: onPointerOut,
            onMouseEnter: onPointerEnter,
            onMouseLeave: onPointerLeave,
        };
    }
    var pointer = useNativePointerEvents(pointerEvents, ref, options);
    var touch = useTouchEvents(touchEvents, ref, options);
    var mouse = useMouseEvents(mouseEvents, ref, options);
    return mergeUseEventResults(pointer, touch, mouse);
};
/**
 * A Conditional version of usePointerEvents
 * As opposed to usePointerEvents, this function does not require to pass a target or ref.
 * When calling this without a target, the handlers passed to the function are returned
 * @param eventHandlers
 * @param ref
 * @param options
 */
var useConditionalPointerEvents = function (eventHandlers, ref, options) {
    var pointerEventsRef;
    var pointerEventsHandlers;
    var emptyRef = useRef(null);
    // Because hooks can't be conditional, we need to this dance, instead of just an early return
    if (ref) {
        pointerEventsRef = ref;
        pointerEventsHandlers = eventHandlers;
    }
    else {
        pointerEventsRef = emptyRef;
        pointerEventsHandlers = {};
    }
    usePointerEvents(pointerEventsHandlers, pointerEventsRef, options);
    if (ref) {
        return undefined;
    }
    else {
        return eventHandlers;
    }
};

/** @public */
var Point;
(function (Point) {
    /** @beta */
    Point.subtract = function (a, b) {
        return { x: a.x - b.x, y: a.y - b.y };
    };
    /** @beta */
    Point.relativeTo = function (idOrElem) {
        var elem;
        var getElem = function () {
            // Caching element here could be leaky because of React lifecycle
            if (elem !== undefined)
                return elem;
            if (typeof idOrElem === "string") {
                elem = document.getElementById(idOrElem);
            }
            else {
                elem = idOrElem;
            }
            return elem;
        };
        return function (_a) {
            var x = _a.x, y = _a.y;
            var localElem = getElem();
            if (!localElem)
                return undefined;
            var rect = localElem.getBoundingClientRect();
            return {
                x: x - rect.left - window.scrollX,
                y: y - rect.top - window.scrollY,
            };
        };
    };
})(Point || (Point = {}));

/**
 * @internal
 */
var MotionPluginContext = createContext({});
/**
 * @internal
 * @internalremarks For now I think this should remain a private API for our own use
 * until we can figure out a nicer way of allowing people to add these
 */
function MotionPlugins(_a) {
    var children = _a.children, props = __rest(_a, ["children"]);
    var pluginContext = useContext(MotionPluginContext);
    var value = useRef(__assign({}, pluginContext)).current;
    // Mutative to prevent triggering rerenders in all listening
    // components every time this component renders
    for (var key in props) {
        value[key] = props[key];
    }
    return (createElement(MotionPluginContext.Provider, { value: value }, children));
}

function startDevicePoint(session) {
    return session.pointHistory[0];
}
function lastDevicePoint(session) {
    return session.pointHistory[session.pointHistory.length - 1];
}
function getVelocity$1(session, timeDelta) {
    var pointHistory = session.pointHistory;
    if (pointHistory.length < 2) {
        return { x: 0, y: 0 };
    }
    var i = pointHistory.length - 1;
    var timestampedPoint = null;
    var lastPoint = lastDevicePoint(session);
    while (i >= 0) {
        timestampedPoint = pointHistory[i];
        if (lastPoint.timestamp - timestampedPoint.timestamp >
            timeDelta * 1000) {
            break;
        }
        i--;
    }
    if (!timestampedPoint) {
        return { x: 0, y: 0 };
    }
    var time = (lastPoint.timestamp - timestampedPoint.timestamp) / 1000;
    if (time === 0) {
        return { x: 0, y: 0 };
    }
    var currentVelocity = {
        x: (lastPoint.x - timestampedPoint.x) / time,
        y: (lastPoint.y - timestampedPoint.y) / time,
    };
    if (currentVelocity.x === Infinity) {
        currentVelocity.x = 0;
    }
    if (currentVelocity.y === Infinity) {
        currentVelocity.y = 0;
    }
    return currentVelocity;
}
function usePanGesture(_a, ref) {
    var onPan = _a.onPan, onPanStart = _a.onPanStart, onPanEnd = _a.onPanEnd;
    var session = useRef(null);
    var pointer = useRef(null);
    var lastMoveEvent = useRef(null);
    var lastMoveEventInfo = useRef(null);
    var transformPagePoint = useContext(MotionPluginContext).transformPagePoint;
    var updatePoint = useCallback(function () {
        if (!session.current ||
            pointer.current === null ||
            lastMoveEventInfo.current === null ||
            lastMoveEvent.current === null) {
            console.error("Pointer move without started session"); // eslint-disable-line no-console
            return;
        }
        var point = lastMoveEventInfo.current.point;
        var delta = Point.subtract(point, lastDevicePoint(session.current));
        var offset = Point.subtract(point, startDevicePoint(session.current));
        var timestamp = getFrameData().timestamp;
        session.current.pointHistory.push(__assign({}, point, { timestamp: timestamp }));
        pointer.current.x.set(point.x);
        pointer.current.y.set(point.y);
        if (Math.abs(delta.x) > 0 || Math.abs(delta.y) > 0) {
            var velocity = getVelocity$1(session.current, 0.1);
            var info = {
                point: point,
                delta: delta,
                offset: offset,
                velocity: velocity,
            };
            if (session.current.startEvent) {
                if (onPan) {
                    onPan(lastMoveEvent.current, info);
                }
            }
            else {
                if (onPanStart) {
                    onPanStart(lastMoveEvent.current, info);
                }
                session.current.startEvent = lastMoveEvent.current;
            }
        }
        // Reset delta
        delta.x = 0;
        delta.y = 0;
    }, [onPan, onPanStart]);
    var onPointerMove = useCallback(function (event, info) {
        lastMoveEvent.current = event;
        if (transformPagePoint) {
            lastMoveEventInfo.current = {
                point: transformPagePoint(info.point),
            };
        }
        else {
            lastMoveEventInfo.current = info;
        }
        // Throttle mouse move event to once per frame
        sync.update(updatePoint, true);
    }, [onPan, onPanStart]);
    var onPointerUp = useCallback(function (event, _a) {
        var point = _a.point;
        cancelSync.update(updatePoint);
        if (!session.current || pointer.current === null) {
            console.error("Pointer end without started session"); // eslint-disable-line no-console
            return;
        }
        var delta = Point.subtract(point, lastDevicePoint(session.current));
        var offset = Point.subtract(point, startDevicePoint(session.current));
        var velocity = getVelocity$1(session.current, 0.1);
        stopPointerMove();
        stopPointerUp();
        if (onPanEnd) {
            onPanEnd(event, {
                point: point,
                delta: delta,
                offset: offset,
                velocity: velocity,
            });
        }
        session.current = null;
    }, [onPanEnd, onPointerMove]);
    var _b = usePointerEvents({ onPointerUp: onPointerUp }, safeWindow), startPointerUp = _b[0], stopPointerUp = _b[1];
    var _c = usePointerEvents({ onPointerMove: onPointerMove }, safeWindow, { capture: true }), startPointerMove = _c[0], stopPointerMove = _c[1];
    var onPointerDown = useCallback(function (event, _a) {
        var point = _a.point;
        var initialPoint = transformPagePoint
            ? transformPagePoint(point)
            : point;
        pointer.current = {
            x: motionValue(initialPoint.x),
            y: motionValue(initialPoint.y),
        };
        var timestamp = getFrameData().timestamp;
        session.current = {
            target: event.target,
            pointHistory: [__assign({}, initialPoint, { timestamp: timestamp })],
        };
        startPointerMove();
        startPointerUp();
    }, [onPointerUp, onPointerMove]);
    useEffect(function () {
        return function () {
            stopPointerMove();
            stopPointerUp();
        };
    }, []);
    var handlers = { onPointerDown: onPointerDown };
    if (!onPan && !onPanStart && !onPanEnd) {
        handlers = {};
    }
    return useConditionalPointerEvents(handlers, ref);
}

/**
 * Recursively traverse up the tree to check whether the provided child node
 * is the parent or a descendant of it.
 *
 * @param parent - Element to find
 * @param child - Element to test against parent
 */
var isNodeOrChild = function (parent, child) {
    if (!child) {
        return false;
    }
    else if (parent === child) {
        return true;
    }
    else {
        return isNodeOrChild(parent, child.parentElement);
    }
};

var order = ["whileHover", "whileTap", "whileDrag"];
var getGesturePriority = function (gesture) {
    return order.indexOf(gesture) + 1;
};

function createLock(name) {
    var lock = null;
    return function () {
        var openLock = function () {
            lock = null;
        };
        if (lock === null) {
            lock = name;
            return openLock;
        }
        return false;
    };
}
var globalHorizontalLock = createLock("dragHorizontal");
var globalVerticalLock = createLock("dragVertical");
function getGlobalLock(drag) {
    var lock = false;
    if (drag === "y") {
        lock = globalVerticalLock();
    }
    else if (drag === "x") {
        lock = globalHorizontalLock();
    }
    else {
        var openHorizontal_1 = globalHorizontalLock();
        var openVertical_1 = globalVerticalLock();
        if (openHorizontal_1 && openVertical_1) {
            lock = function () {
                openHorizontal_1();
                openVertical_1();
            };
        }
        else {
            // Release the locks because we don't use them
            if (openHorizontal_1)
                openHorizontal_1();
            if (openVertical_1)
                openVertical_1();
        }
    }
    return lock;
}

var tapGesturePriority = getGesturePriority("whileTap");
/**
 * Explanation:
 *
 * Currently there's a problem where props can change while a gesture is active. In
 * this hook, we have event handlers that are created in `useMemo`, which is sync.
 * The event handlers themselves are activated in `useEvent` with `useEffect`, which
 * is async.
 *
 * Most event handlers are being set like this: onTap={() => {}}
 * This creates a new function every render, thus breaking `useMemo`. When any props
 * are set during a gesture, this creates a new event. We then have a `useEffect` in this
 * hook that, when props change, unsets the previous event handlers. I suspect that at
 * some point, the pointer up event is fired in between the previous one
 * being deactivate and the new one being activated.
 *
 * I attempted to rewrite the hook without `useMemo` at all but encountered the same issue.
 * At some point the `pointerUp` event is being unset and not reapplied until after it fires.
 *
 * By setting props to a ref we maintain longer gesture sessions and don't encounter this funny
 * business, but I'd rather a solution that required less code.
 *
 * @param props
 */
var usePropsRef = function (props) {
    var propsRef = useRef(props);
    for (var key in props) {
        propsRef.current[key] = props[key];
    }
    return propsRef.current;
};
function useTapGesture(props, ref) {
    var isTapping = useRef(false);
    var onTap = props.onTap, onTapStart = props.onTapStart, onTapCancel = props.onTapCancel, whileTap = props.whileTap, controls = props.controls;
    var propsRef = usePropsRef(props);
    if (whileTap && controls) {
        controls.setOverride(whileTap, tapGesturePriority);
    }
    var handlers = useMemo(function () {
        if (!onTap && !onTapStart && !onTapCancel && !whileTap) {
            return {
                onPointerUp: function () { },
                onPointerDown: function () { },
            };
        }
        var onPointerUp = function (event, _a) {
            var point = _a.point;
            if (!isTapping.current) {
                return;
            }
            isTapping.current = false;
            if (controls && propsRef.whileTap) {
                controls.clearOverride(tapGesturePriority);
            }
            // Check the gesture lock - if we get it, it means no drag gesture is active
            // and we can safely fire the tap gesture.
            var openGestureLock = getGlobalLock(true);
            if (!openGestureLock) {
                return;
            }
            openGestureLock();
            if (!ref ||
                !ref.current ||
                !isNodeOrChild(ref.current, event.target)) {
                if (propsRef.onTapCancel) {
                    propsRef.onTapCancel(event, { point: point });
                }
                return;
            }
            if (propsRef.onTap) {
                propsRef.onTap(event, { point: point });
            }
        };
        var onPointerDown = function (event, _a) {
            var point = _a.point;
            startPointerUp();
            if (!ref || !ref.current)
                return;
            isTapping.current = true;
            if (propsRef.onTapStart) {
                propsRef.onTapStart(event, { point: point });
            }
            if (controls && propsRef.whileTap) {
                controls.startOverride(tapGesturePriority);
            }
        };
        return { onPointerUp: onPointerUp, onPointerDown: onPointerDown };
    }, [
        onTap === undefined,
        onTapStart === undefined,
        onTapCancel === undefined,
    ]);
    var _a = usePointerEvents({ onPointerUp: handlers.onPointerUp }, safeWindow), startPointerUp = _a[0], stopPointerUp = _a[1];
    useEffect(function () { return function () {
        stopPointerUp();
    }; }, [handlers.onPointerUp]);
    return useConditionalPointerEvents({ onPointerDown: handlers.onPointerDown }, ref);
}

var hoverPriority = getGesturePriority("whileHover");
// TODO: Optimisation here is find a way to conditionally add these listeners based on
// whether we're receiving hover or event listeners
/**
 *
 * @param props
 * @param ref
 * @internal
 */
function useHoverGesture(_a, ref) {
    var whileHover = _a.whileHover, onHoverStart = _a.onHoverStart, onHoverEnd = _a.onHoverEnd, controls = _a.controls;
    if (whileHover && controls) {
        controls.setOverride(whileHover, hoverPriority);
    }
    var handlers = useMemo(function () {
        var onPointerEnter = function (event) {
            if (onHoverStart)
                onHoverStart(event);
            if (whileHover && controls) {
                controls.startOverride(hoverPriority);
            }
        };
        var onPointerLeave = function (event) {
            if (onHoverEnd)
                onHoverEnd(event);
            if (whileHover && controls) {
                controls.clearOverride(hoverPriority);
            }
        };
        return { onPointerEnter: onPointerEnter, onPointerLeave: onPointerLeave };
    }, [whileHover, onHoverStart, onHoverEnd, controls]);
    return useConditionalPointerEvents(handlers, ref);
}

/**
 * Add pan and tap gesture recognition to an element.
 *
 * @param props - Gesture event handlers
 * @param ref - React `ref` containing a DOM `Element`
 * @public
 */
function useGestures(props, ref) {
    usePanGesture(props, ref);
    useTapGesture(props, ref);
    useHoverGesture(props, ref);
}

var flattenConstraints = function (constraints) {
    if (!constraints) {
        return [0, 0, 0, 0];
    }
    else {
        var top_1 = constraints.top, left = constraints.left, bottom = constraints.bottom, right = constraints.right;
        return [top_1, left, bottom, right];
    }
};
function shouldDrag(direction, drag, currentDirection) {
    return ((drag === true || drag === direction) &&
        (currentDirection === null || currentDirection === direction));
}
var getConstraints = function (axis, _a) {
    var top = _a.top, right = _a.right, bottom = _a.bottom, left = _a.left;
    if (axis === "x") {
        return { min: left, max: right };
    }
    else {
        return { min: top, max: bottom };
    }
};
function applyConstraints(axis, value, constraints, dragElastic) {
    var constrainedValue = value instanceof MotionValue ? value.get() : value;
    if (!constraints) {
        return constrainedValue;
    }
    var _a = getConstraints(axis, constraints), min = _a.min, max = _a.max;
    if (min !== undefined && constrainedValue < min) {
        constrainedValue = dragElastic
            ? applyOverdrag(min, constrainedValue, dragElastic)
            : Math.max(min, constrainedValue);
    }
    else if (max !== undefined && constrainedValue > max) {
        constrainedValue = dragElastic
            ? applyOverdrag(max, constrainedValue, dragElastic)
            : Math.min(max, constrainedValue);
    }
    if (value instanceof MotionValue) {
        value.set(constrainedValue);
    }
    return constrainedValue;
}
var applyOverdrag = function (origin, current, dragElastic) {
    var dragFactor = typeof dragElastic === "number" ? dragElastic : 0.5;
    return mix(origin, current, dragFactor);
};
/**
 * A hook that allows an element to be dragged.
 *
 * @internalremarks
 * TODO:
 * 1. Allow `dragMomentum` to accept richer properties that adjust this behaviour
 * 2. Allow a parent motion component to become the drag boundaries for a child
 *
 * @param param
 * @param ref
 * @param values
 * @param controls
 *
 * @internal
 */
function useDraggable(_a, ref, values, controls) {
    var _b = _a.drag, drag = _b === void 0 ? false : _b, _c = _a.dragDirectionLock, dragDirectionLock = _c === void 0 ? false : _c, _d = _a.dragPropagation, dragPropagation = _d === void 0 ? false : _d, _e = _a.dragConstraints, dragConstraints = _e === void 0 ? false : _e, _f = _a.dragElastic, dragElastic = _f === void 0 ? true : _f, _g = _a.dragMomentum, dragMomentum = _g === void 0 ? true : _g, dragTransition = _a.dragTransition, onDragStart = _a.onDragStart, onDragEnd = _a.onDragEnd, onDrag = _a.onDrag, onDirectionLock = _a.onDirectionLock, onDragTransitionEnd = _a.onDragTransitionEnd;
    var point = useRef({}).current;
    var onDragRef = useRef(onDrag);
    var origin = useRef({ x: 0, y: 0 }).current;
    onDragRef.current = onDrag;
    var handlers = useMemo(function () {
        if (!drag)
            return {};
        var currentDirection = null;
        var openGlobalLock = null;
        if (shouldDrag("x", drag, currentDirection)) {
            var x = values.get("x", 0);
            applyConstraints("x", x, dragConstraints, dragElastic);
            point.x = x;
        }
        if (shouldDrag("y", drag, currentDirection)) {
            var y = values.get("y", 0);
            applyConstraints("y", y, dragConstraints, dragElastic);
            point.y = y;
        }
        var updatePoint = function (axis, offset) {
            var p = point[axis];
            if (!shouldDrag(axis, drag, currentDirection) || !p) {
                return;
            }
            var current = origin[axis] + offset[axis];
            current = applyConstraints(axis, current, dragConstraints, dragElastic);
            p.set(current);
        };
        var onPointerDown = function () {
            if (point.x)
                point.x.stop();
            if (point.y)
                point.y.stop();
        };
        var onPanStart = function (event, info) {
            var handle = function (axis) {
                var axisPoint = point[axis];
                if (!axisPoint)
                    return;
                origin[axis] = axisPoint.get();
                axisPoint.stop();
            };
            handle("x");
            handle("y");
            if (!dragPropagation) {
                openGlobalLock = getGlobalLock(drag);
                if (!openGlobalLock) {
                    return;
                }
            }
            currentDirection = null;
            onDragStart && onDragStart(event, info);
        };
        var onPan = function (event, info) {
            if (!dragPropagation && !openGlobalLock) {
                return;
            }
            var offset = info.offset;
            if (dragDirectionLock) {
                if (currentDirection === null) {
                    currentDirection = getCurrentDirection(offset);
                    if (currentDirection !== null) {
                        onDirectionLock && onDirectionLock(currentDirection);
                    }
                    return;
                }
            }
            updatePoint("x", offset);
            updatePoint("y", offset);
            // here we use ref to call only the last event handler
            if (onDragRef.current) {
                onDragRef.current(event, __assign({}, info, { point: {
                        x: point.x ? point.x.get() : 0,
                        y: point.y ? point.y.get() : 0,
                    } }));
            }
        };
        var onPanEnd = function (event, info) {
            var velocity = info.velocity;
            if (!dragPropagation && openGlobalLock) {
                openGlobalLock();
            }
            else if (!openGlobalLock) {
                return;
            }
            if (dragMomentum) {
                var startMomentum = function (axis) {
                    var _a;
                    if (!shouldDrag(axis, drag, currentDirection)) {
                        return;
                    }
                    var transition = dragConstraints
                        ? getConstraints(axis, dragConstraints)
                        : {};
                    return controls.start((_a = {},
                        _a[axis] = 0,
                        _a.transition = __assign({ type: "inertia", velocity: velocity[axis], bounceStiffness: 200, bounceDamping: 40, timeConstant: 750, restDelta: 1 }, dragTransition, transition),
                        _a));
                };
                Promise.all([startMomentum("x"), startMomentum("y")]).then(function () {
                    onDragTransitionEnd && onDragTransitionEnd();
                });
            }
            onDragEnd && onDragEnd(event, info);
        };
        return {
            onPanStart: onPanStart,
            onPan: onPan,
            onPanEnd: onPanEnd,
            onPointerDown: onPointerDown,
        };
    }, [drag].concat(flattenConstraints(dragConstraints), [onDragTransitionEnd]));
    usePanGesture(handlers, ref);
    usePointerEvents({ onPointerDown: handlers.onPointerDown }, ref);
}
function getCurrentDirection(offset) {
    var lockThreshold = 10;
    var direction = null;
    if (Math.abs(offset.y) > lockThreshold) {
        direction = "y";
    }
    else if (Math.abs(offset.x) > lockThreshold) {
        direction = "x";
    }
    return direction;
}

/**
 * Handle the `animate` prop when its an object of values, ie:
 *
 * ```jsx
 * <motion.div animate={{ opacity: 1 }} />
 * ```
 *
 * @internalremarks
 * It might be worth consolidating this with `use-variants`
 *
 * ```jsx
 * <motion.div animate="visible" />
 * ```
 *
 * @param target
 * @param controls
 * @param values
 * @param transition
 *
 * @internal
 */
function useAnimateProp(target, controls, values, transition) {
    var isInitialRender = useRef(true);
    var prevValues = useRef(target);
    useEffect(function () {
        var toAnimate = {};
        for (var key in prevValues.current) {
            var hasUpdated = target[key] !== undefined &&
                prevValues.current[key] !== target[key];
            var shouldAnimateOnMount = isInitialRender.current &&
                (!values.has(key) || values.get(key) !== target[key]);
            if (hasUpdated || shouldAnimateOnMount) {
                toAnimate[key] = target[key];
            }
        }
        isInitialRender.current = false;
        prevValues.current = __assign({}, prevValues.current, target);
        if (Object.keys(toAnimate).length) {
            controls.start(toAnimate, transition);
        }
    }, [target]);
}

var AnimatePropType;
(function (AnimatePropType) {
    AnimatePropType["Target"] = "Target";
    AnimatePropType["VariantLabel"] = "VariantLabel";
    AnimatePropType["AnimationSubscription"] = "AnimationSubscription";
})(AnimatePropType || (AnimatePropType = {}));

var _a, _b;
var makeHookComponent = function (hook) { return function (props) {
    hook(props);
    return null;
}; };
var AnimatePropComponents = (_a = {},
    _a[AnimatePropType.Target] = makeHookComponent(function (_a) {
        var animate = _a.animate, controls = _a.controls, values = _a.values, transition = _a.transition;
        return useAnimateProp(animate, controls, values, transition);
    }),
    _a[AnimatePropType.VariantLabel] = makeHookComponent(function (_a) {
        var animate = _a.animate, _b = _a.inherit, inherit = _b === void 0 ? true : _b, controls = _a.controls, initial = _a.initial;
        return useVariants(animate, inherit, controls, initial);
    }),
    _a[AnimatePropType.AnimationSubscription] = makeHookComponent(function (_a) {
        var animate = _a.animate, controls = _a.controls;
        return useAnimationGroupSubscription(animate, controls);
    }),
    _a);
var isVariantLabel = function (prop) {
    return Array.isArray(prop) || typeof prop === "string";
};
var isAnimationSubscription = function (_a) {
    var animate = _a.animate;
    return animate instanceof AnimationControls;
};
var animationProps = ["initial", "animate", "whileTap", "whileHover"];
var animatePropTypeTests = (_b = {},
    _b[AnimatePropType.Target] = function (props) {
        return (props.animate !== undefined &&
            !isVariantLabel(props.animate) &&
            !isAnimationSubscription(props));
    },
    _b[AnimatePropType.VariantLabel] = function (props) {
        return (props.variants !== undefined ||
            animationProps.some(function (key) { return typeof props[key] === "string"; }));
    },
    _b[AnimatePropType.AnimationSubscription] = isAnimationSubscription,
    _b);
var getAnimatePropType = function (props) {
    for (var key in AnimatePropType) {
        if (animatePropTypeTests[key](props))
            return key;
    }
    return undefined;
};
var getAnimateComponent = function (props, isStatic) {
    if (isStatic === void 0) { isStatic = false; }
    var animatePropType = getAnimatePropType(props);
    return !isStatic && animatePropType
        ? AnimatePropComponents[animatePropType]
        : undefined;
};
var gestureProps = [
    "drag",
    "onPan",
    "onPanStart",
    "onPanEnd",
    "onTap",
    "onTapStart",
    "onTapCancel",
    "whileTap",
    "whileHover",
    "onHoverStart",
    "onHoverEnd",
];
var isGesturesEnabled = function (props) {
    return gestureProps.some(function (key) { return props.hasOwnProperty(key); });
};
var isDragEnabled = function (props) { return !!props.drag; };
var Gestures = makeHookComponent(function (_a) {
    var innerRef = _a.innerRef, props = __rest(_a, ["innerRef"]);
    return useGestures(props, innerRef);
});
var Draggable = makeHookComponent(function (_a) {
    var innerRef = _a.innerRef, values = _a.values, controls = _a.controls, props = __rest(_a, ["innerRef", "values", "controls"]);
    return useDraggable(props, innerRef, values, controls);
});
var eventHandlers = new Set([
    "onAnimationComplete",
    "onUpdate",
    "onDragStart",
    "onDrag",
    "onDragEnd",
    "onDirectionLock",
    "onDragTransitionEnd"
].concat(gestureProps));
var validProps = function (props) {
    var valid = {};
    for (var key in props) {
        if (isPropValid(key) && !eventHandlers.has(key)) {
            valid[key] = props[key];
        }
    }
    return valid;
};
var RenderComponent = function (_a) {
    var base = _a.base, props = _a.props, innerRef = _a.innerRef, style = _a.style, values = _a.values, isStatic = _a.isStatic;
    var isDOM = typeof base === "string";
    var isSVG = isDOM && svgElements.indexOf(base) !== -1;
    var forwardProps = isDOM ? validProps(props) : props;
    return createElement(base, __assign({}, forwardProps, { ref: innerRef, style: isSVG ? style : buildStyleAttr(values, style, isStatic) }));
};
var checkShouldInheritVariant = function (_a) {
    var animate = _a.animate, _b = _a.inherit, inherit = _b === void 0 ? true : _b, variants = _a.variants, whileHover = _a.whileHover, whileTap = _a.whileTap;
    var isVariantChild = inherit && variants && !animate && !whileHover && !whileTap;
    var isAnimationHookChild = inherit && animate instanceof AnimationControls;
    if (isVariantChild || isAnimationHookChild) {
        return true;
    }
    else {
        return false;
    }
};

/**
 * @internal
 */
var createMotionComponent = function (Component) {
    function MotionComponent(props, externalRef) {
        var ref = useExternalRef(externalRef);
        var parentContext = useContext(MotionContext);
        var isStatic = parentContext.static || props.static || false;
        var values = useMotionValues(props, isStatic);
        var style = useMotionStyles(values, props.style, props.transformValues);
        var shouldInheritVariant = checkShouldInheritVariant(props);
        var controls = useComponentAnimationControls(values, props, ref, shouldInheritVariant);
        var context = useMotionContext(parentContext, controls, isStatic, props.initial);
        // Add functionality
        var Animate = getAnimateComponent(props, context.static);
        var handleAnimate = Animate && (createElement(Animate, __assign({}, props, { inherit: shouldInheritVariant, innerRef: ref, values: values, controls: controls })));
        var handleGestures = !context.static && isGesturesEnabled(props) && (createElement(Gestures, __assign({}, props, { values: values, controls: controls, innerRef: ref })));
        var handleDrag = !context.static && isDragEnabled(props) && (createElement(Draggable, __assign({}, props, { innerRef: ref, controls: controls, values: values })));
        // We use an intermediate component here rather than calling `createElement` directly
        // because we want to resolve the style from our motion values only once every
        // functional component has resolved. Resolving it here would do it before the functional components
        // themselves are executed.
        var handleComponent = (createElement(RenderComponent, { base: Component, props: props, innerRef: ref, style: style, values: values, isStatic: isStatic }));
        return (createElement(MotionContext.Provider, { value: context },
            createElement(MountMotionValues, { ref: ref, values: values }),
            handleAnimate,
            handleGestures,
            handleDrag,
            handleComponent));
    }
    return forwardRef(MotionComponent);
};

/**
 * HTML & SVG components, optimised for use with gestures and animation. These can be used as
 * drop-in replacements for any HTML & SVG component, all CSS & SVG properties are supported.
 *
 * ```jsx
 * <motion.div animate={{ x: 100 }} />
 *
 * <motion.p animate={{ height: 200 }} />
 *
 * <svg><motion.circle r={10} animate={{ r: 20 }} /></svg>
 * ```
 *
 * @public
 */
var motion = elements.reduce(function (acc, element) {
    acc[element] = createMotionComponent(element);
    return acc;
}, {
    custom: createMotionComponent,
});

/**
 * Creates a `MotionValue` to track the state and velocity of a value.
 *
 * Usually, these are created automatically. For advanced use-cases, like use with `useTransform`, you can create `MotionValue`s externally and pass them into the animated component via the `style` prop.
 *
 * ```jsx
 * export function MyComponent() {
 *   const scale = useMotionValue(1)
 *
 *   return <Frame style={{ scale: scale }} />
 * }
 * ```
 *
 * @param initial - The initial state.
 *
 * @public
 */
function useMotionValue(initial) {
    return useMemo(function () { return motionValue(initial); }, []);
}

/**
 * If the provided value is a MotionValue, this returns the actual value, otherwise just the value itself
 *
 * TODO: Remove and move to library
 *
 * @internal
 */
function unwrapMotionValue(value) {
    var unwrappedValue = value instanceof MotionValue ? value.get() : value;
    return isCustomValue(unwrappedValue)
        ? unwrappedValue.toValue()
        : unwrappedValue;
}

function useCustomValue(parent, getTransformer, comparitor) {
    var value = useRef(null);
    return useMemo(function () {
        if (value.current)
            value.current.destroy();
        value.current = parent.addChild({ transformer: getTransformer() });
        return value.current;
    }, comparitor);
}

var isCustomValueType = function (v) {
    return typeof v === "object" && v.mix;
};
var getMixer = function (v) { return (isCustomValueType(v) ? v.mix : undefined); };
function transform() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var useImmediate = !Array.isArray(args[0]);
    var argOffset = useImmediate ? 0 : -1;
    var inputValue = args[0 + argOffset];
    var inputRange = args[1 + argOffset];
    var outputRange = args[2 + argOffset];
    var options = args[3 + argOffset];
    var interpolator = interpolate(inputRange, outputRange, __assign({ mixer: getMixer(outputRange[0]) }, options));
    return useImmediate ? interpolator(inputValue) : interpolator;
}

var isTransformer = function (v) {
    return typeof v === "function";
};
var noop$1 = function () { return function (v) { return v; }; };
function useTransform(value, customTransform, to, options) {
    var comparitor = [value];
    var transformer = noop$1;
    if (isTransformer(customTransform)) {
        transformer = function () { return customTransform; };
    }
    else if (Array.isArray(to)) {
        var from_1 = customTransform;
        transformer = function () { return transform(from_1, to, options); };
        comparitor = [value, from_1.join(","), to.join(",")];
    }
    return useCustomValue(value, transformer, comparitor);
}

var scrollX = motionValue(0);
var scrollY = motionValue(0);
var scrollXProgress = motionValue(0);
var scrollYProgress = motionValue(0);
var hasEventListener = false;
var addScrollListener = function () {
    hasEventListener = true;
    if (typeof window === "undefined")
        return;
    var updateScrollValues = function () {
        var xOffset = window.pageXOffset;
        var yOffset = window.pageYOffset;
        // Set absolute positions
        scrollX.set(xOffset);
        scrollY.set(yOffset);
        // Set progress
        var maxXOffset = document.body.clientWidth - window.innerWidth;
        scrollXProgress.set(xOffset === 0 ? 0 : xOffset / maxXOffset);
        var maxYOffset = document.body.clientHeight - window.innerHeight;
        scrollYProgress.set(yOffset === 0 ? 0 : yOffset / maxYOffset);
    };
    updateScrollValues();
    window.addEventListener("resize", updateScrollValues);
    window.addEventListener("scroll", updateScrollValues, { passive: true });
};
var viewportMotionValues = {
    scrollX: scrollX,
    scrollY: scrollY,
    scrollXProgress: scrollXProgress,
    scrollYProgress: scrollYProgress,
};
/**
 * Provides a `MotionValue` that updates when the viewport scrolls.
 * Returns the following four values.
 *
 * - `scrollX` — Horizontal scroll distance in pixels.
 * - `scrollY` — Vertical scroll distance in pixels.
 * - `scrollXProgress` — Horizontal scroll distance between `0` and `1`.
 * - `scrollYProgress` — Vertical scroll distance between `0` and `1`.
 *
 *
 * ```jsx
 * import * as React from "react"
 * import {
 *   Frame,
 *   useViewportScrollValues,
 *   useTransform
 * } from "framer"
 *
 * export function MyComponent() {
 *   const { scrollYProgress } = useViewportScrollValues()
 *   const scaleX = useTransform(scrollYProgress, [0, 1], [0, 1])
 *
 *   return <Frame style={{ scaleX }} />
 * }
 * ```
 *
 * @public
 */
function useViewportScrollValues() {
    if (!hasEventListener) {
        addScrollListener();
    }
    return viewportMotionValues;
}

/**
 * The `useAnimation` hook returns `AnimationControls`, which can be used to manually start, stop and sequence animations on one or more Frames.
 *
 * The returned `AnimationControls` should be passed as the `animate` property of the `Frame`. It can be passed to any number of Frames. These Frames can be animated with the `start` method. It accepts the same properties as the `animate` property.
 *
 * ```jsx
 *
 * import * as React from "react"
 * import { Frame, useAnimation } from "framer"
 *
 * export function MyComponent(props) {
 *    const animation = useAnimation()
 *
 *    animation.start({
 *        x: 100,
 *        transition: { duration: 0.5 },
 *    })
 *
 *    return <Frame animate={animation} />
 * }
 *
 * ```
 *
 * @param variants - An optional named map of variants.
 * @param defaultTransition - An optional default transition to use when a variant doesn’t have an explicit `transition` property set.
 * @returns Animation controller with `start` and `stop` methods
 *
 * @public
 */
function useAnimation(variants, defaultTransition) {
    var animationManager = useMemo(function () { return new AnimationControls(); }, []);
    if (variants) {
        animationManager.setVariants(variants);
    }
    if (defaultTransition) {
        animationManager.setDefaultTransition(defaultTransition);
    }
    useEffect(function () {
        animationManager.mount();
        return function () { return animationManager.unmount(); };
    }, []);
    return animationManager;
}

/**
 * Cycles through a series of visual properties. Can be used to toggle between or cycle through animations. It works similar to `useState` in React. It is provided an initial array of possible states, and returns an array of two arguments.
 *
 *
 * ```jsx
 * import * as React from "react"
 * import { Frame, useCycle } from "framer"
 *
 * export function MyComponent() {
 *   const [x, cycleX] = useCycle(0, 50, 100)
 *
 *   return (
 *     <Frame
 *       animate={{ x: x }}
 *       onTap={() => cycleX()}
 *      />
 *    )
 * }
 * ```
 *
 * @param items... - items to cycle through
 * @returns [currentState, cycleState]
 *
 * @public
 */
function useCycle() {
    var items = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        items[_i] = arguments[_i];
    }
    // TODO: After Framer X beta, remove this warning
    warning(items.length > 1, "useCycle syntax has changed. `useCycle([0, 1, 2])` becomes `useCycle(0, 1, 2)`");
    var index = useRef(0);
    var _a = useState(items[index.current]), item = _a[0], setItem = _a[1];
    return [
        item,
        function (next) {
            index.current =
                typeof next !== "number"
                    ? wrap(0, items.length, index.current + 1)
                    : next;
            setItem(items[index.current]);
        },
    ];
}

export { motion, useExternalRef, createMotionComponent, htmlElements, svgElements, useMotionValue, MotionValue, motionValue, unwrapMotionValue, useTransform, useViewportScrollValues, AnimationControls, animationControls, useAnimation, useTapGesture, usePanGesture, useGestures, useCycle, transform, MotionPlugins, MotionPluginContext, MotionContext, Point, safeWindow };
